@@ -9,111 +9,54 @@ use Exporter;
 use base 'Exporter';
 use vars qw(@EXPORT $VERSION $AUTHORITY);
 
-# This is the mother class for each class that uses Coat
+use Coat::Meta;
 use Coat::Object;
+use Coat::Types;
 
-$VERSION   = '0.1_0.3';
+$VERSION   = '0.1_0.4';
 $AUTHORITY = 'cpan:SUKRIA';
 
 # our exported keywords for class description
 @EXPORT = qw(has extends before after around);
 
-# This is the class placeholder for attribute descriptions
-# it's present in scope of the class itself, not for each instance
-my $CLASS_ATTRS = {};
+# Prototypes for private methods
+sub _bind_coderef_to_symbol($$);
+sub _extends_class($;$);
+sub _value_is_valid($$);
+sub _compile_around_modifier(@);
+sub _build_sub_with_hook($$);
 
-# local accessors for class attributes/descriptions
-
-# declare/get a class description
-sub class { $CLASS_ATTRS->{ $_[0] } ||= {} }
-
-# set/get an attribute of a class
-sub class_attr {
-    @_ == 3
-      ? $CLASS_ATTRS->{ $_[0] }{ $_[1] } = $_[2]
-      : $CLASS_ATTRS->{ $_[0] }{ $_[1] } ||= {};
-}
-
-sub class_exists     { exists $CLASS_ATTRS->{ $_[0] } }
-sub class_has_attr   { exists $CLASS_ATTRS->{ $_[0] }{ $_[1] } }
-sub class_set_mother { $CLASS_ATTRS->{__mother}{ $_[0] } = $_[1] }
-sub class_get_mother { $CLASS_ATTRS->{__mother}{ $_[0] } }
-
-# hooks for a module
-sub hooks        { $CLASS_ATTRS->{__hooks}{ $_[0] } }
-sub hooks_before { $CLASS_ATTRS->{__hooks}{ $_[0] }{before}{ $_[1] } ||= [] }
-sub hooks_after  { $CLASS_ATTRS->{__hooks}{ $_[0] }{after}{ $_[1] } ||= [] }
-sub hooks_around { $CLASS_ATTRS->{__hooks}{ $_[0] }{around}{ $_[1] } ||= [] }
-
-# helper for copying a class description (used when inheriting from a class)
-sub __copy_class_description($$) {
-    my ( $source, $dest ) = @_;
-    foreach my $key ( keys %{ $CLASS_ATTRS->{$source} } ) {
-        $CLASS_ATTRS->{$dest}{$key} = $CLASS_ATTRS->{$source}{$key};
-    }
-}
+##############################################################################
+# Public static methods
+##############################################################################
 
 # has() declares an attribute and builds the corresponding accessors
 sub has {
-    my ( $name, %options ) = @_;
-    confess "Attribute is a reference, cannot declare" if ref($name);
+    my ( $attribute, %options ) = @_;
+    confess "Attribute is a reference, cannot declare" if ref($attribute);
 
-    my $scope    = getscope();
-    my $accessor = "${scope}::${name}";
+    my $class    = getscope();
+    my $accessor = "${class}::${attribute}";
 
-    class_attr( $scope, $name, { type => 'Scalar', %options } );
+    my $attr = Coat::Meta->attribute( $class, $attribute, \%options);
 
     my $accessor_code = sub {
         my ( $self, $value ) = @_;
-        confess "Unknown attribute '$name' for class " . ref($self)
-          unless $self->has_attr($name);
-
+        
         # want a set()
         if ( @_ > 1 ) {
-            my $attrs = $self->meta;
-            my $type  = $attrs->{$name}{type};
-
-            # FIXME : this will be better when we have Coat::Types implemented
-            confess "$type '$name' cannot be set to '$value'"
-              unless ( __value_is_valid( $value, $type ) );
-
-            $self->{_values}{$name} = $value;
-            return $value;
+            Coat::Types->validate( $attr->{'isa'}, $attribute, $value );
+            return $self->{$attribute} = $value;
         }
 
         # want a get()
         else {
-            return $self->{_values}{$name};
+            return $self->{$attribute};
         }
     };
 
     # now bind the subref to the appropriate symbol in the caller class
-    __bind_coderef_to_symbol( $accessor_code, $accessor );
-}
-
-# the private method for declaring inheritance, we can here overide the
-# caller class with a random one, useful for our internal cooking, see import().
-sub __extends_class {
-    my ( $mothers, $class ) = @_;
-    $class = getscope() unless defined $class;
-
-    # the extends mechanism overwrite the @ISA list
-    { no strict 'refs'; @{"${class}::ISA"} = (); }
-
-    # then we inherit from all the mothers given
-    foreach my $mother (@$mothers) {
-        confess "Class '$mother' is unknown, cannot extends"
-          unless class_exists($mother);
-
-        # first we inherit the class description from our mother
-        __copy_class_description( $mother, $class );
-
-        # add the mother to our ancestors
-        { no strict 'refs'; push @{"${class}::ISA"}, $mother; }
-
-        # save the fact that $class inherits from $mother
-        class_set_mother( $class, $mother );
-    }
+    _bind_coderef_to_symbol( $accessor_code, $accessor );
 }
 
 # the public inheritance method, takes a list of class we should inherit from
@@ -121,79 +64,7 @@ sub extends {
     my (@mothers) = @_;
     confess "Cannot extend without a class name"
       unless @mothers;
-    __extends_class( \@mothers, getscope() );
-}
-
-# returns the parent class of the class given
-sub super {
-    my ($class) = @_;
-    $class = getscope() unless defined $class;
-    return class_get_mother($class);
-}
-
-# local helpers for building wrapped methods
-sub __hooks_before_push { push @{ hooks_before( $_[0], $_[1] ) }, $_[2] }
-sub __hooks_after_push { push @{ hooks_after( $_[0], $_[1] ) }, $_[2] }
-sub __hooks_around_push { push @{ hooks_around( $_[0], $_[1] ) }, $_[2] }
-
-# The idea here is to loop on each coderef given
-# and build subs to ensure the orig coderef is correctly propagated.
-# -> We rewrite the "around" hooks defined to pass their coderef neighboor as
-# a first argument.
-# (big thank to STEVAN's Class::MOP here, which was helpful with the idea of
-# $compile_around_method)
-sub __compile_around_modifier {
-    {
-        my $orig = shift;
-        return $orig unless @_;
-
-        my $hook = shift;
-        @_ = ( sub { $hook->( $orig, @_ ) }, @_ );
-        redo;
-    }
-}
-
-# This one is the wrapper builder for method with hooks.
-# It can mix up before, after and around hooks.
-sub __build_sub_with_hook($$) {
-    my ( $class, $method ) = @_;
-
-    my $super        = super($class);
-    my $full_method  = "${class}::${method}";
-    my $super_method = *{ qualify_to_ref( $method => $super ) };
-
-    my ( $before, $after, $around ) = (
-        hooks_before( $class, $method ),
-        hooks_after( $class, $method ),
-        hooks_around( $class, $method )
-    );
-
-    my $modified_method_code = sub {
-        my ( $self, @args ) = @_;
-        my @result;
-        my $result;
-
-        $_->(@_) for @$before;
-
-        my $around_modifier =
-          __compile_around_modifier( \&$super_method, @$around );
-
-        ( defined wantarray )
-          ? (
-            wantarray
-            ? ( @result = $around_modifier->(@_) )
-            : ( $result = $around_modifier->(@_) )
-          )
-          : ( $around_modifier->(@_) );
-
-        $_->(@_) for @$after;
-
-        return unless defined wantarray;
-        return wantarray ? @result : $result;
-    };
-
-    # now bind the new method to the appropriate symbol
-    __bind_coderef_to_symbol( $modified_method_code, $full_method );
+    _extends_class( \@mothers, getscope() );
 }
 
 # the before hook catches the call to an inherited method and exectue
@@ -201,8 +72,8 @@ sub __build_sub_with_hook($$) {
 sub before {
     my ( $method, $code ) = @_;
     my $class = getscope();
-    __hooks_before_push( $class, $method, $code );
-    __build_sub_with_hook( $class, $method );
+    Coat::Meta->before_modifiers( $class, $method, $code );
+    _build_sub_with_hook( $class, $method );
 }
 
 # the after hook catches the call to an inherited method and executes
@@ -210,8 +81,8 @@ sub before {
 sub after {
     my ( $method, $code ) = @_;
     my $class = getscope();
-    __hooks_after_push( $class, $method, $code );
-    __build_sub_with_hook( $class, $method );
+    Coat::Meta->after_modifiers( $class, $method, $code );
+    _build_sub_with_hook( $class, $method );
 }
 
 # the around hook catches the call to an inherited method and lets you do
@@ -220,8 +91,8 @@ sub after {
 sub around {
     my ( $method, $code ) = @_;
     my $class = getscope();
-    __hooks_around_push( $class, $method, $code );
-    __build_sub_with_hook( $class, $method );
+    Coat::Meta->around_modifiers( $class, $method, $code );
+    _build_sub_with_hook( $class, $method );
 }
 
 # we override the import method to actually force the "strict" and "warnings"
@@ -234,13 +105,13 @@ sub import {
     warnings->import;
 
     # delcare the class
-    class( getscope() );
+    Coat::Meta->class( getscope() );
 
     # be sure Coat::Object is known as a valid class
-    class('Coat::Object');
+    Coat::Meta->class('Coat::Object');
 
     # force inheritance from Coat::Object
-    __extends_class( ['Coat::Object'], getscope() );
+    _extends_class( ['Coat::Object'], getscope() );
 
     return if $caller eq 'main';
     Coat->export_to_level( 1, @_ );
@@ -264,10 +135,91 @@ sub getscope {
 }
 
 ##############################################################################
-# Private methods (only called from Coat.pm)
+# Private methods
 ##############################################################################
 
-sub __bind_coderef_to_symbol($$) {
+# The idea here is to loop on each coderef given
+# and build subs to ensure the orig coderef is correctly propagated.
+# -> We rewrite the "around" hooks defined to pass their coderef neighboor as
+# a first argument.
+# (big thank to STEVAN's Class::MOP here, which was helpful with the idea of
+# $compile_around_method)
+sub _compile_around_modifier(@) {
+    {
+        my $orig = shift;
+        return $orig unless @_;
+
+        my $hook = shift;
+        @_ = ( sub { $hook->( $orig, @_ ) }, @_ );
+        redo;
+    }
+}
+
+# This one is the wrapper builder for method with hooks.
+# It can mix up before, after and around hooks.
+sub _build_sub_with_hook($$) {
+    my ( $class, $method ) = @_;
+
+    my $parents      = Coat::Meta->parents( $class );
+    # FIXME : we have to find the good super: the one who provides the sub
+    my $super = $parents->[scalar(@$parents) - 1];
+
+    my $full_method  = "${class}::${method}";
+    my $super_method = *{ qualify_to_ref( $method => $super ) };
+
+    my ( $before, $after, $around ) = (
+        Coat::Meta->before_modifiers( $class, $method ),
+        Coat::Meta->after_modifiers ( $class, $method ),
+        Coat::Meta->around_modifiers( $class, $method ),
+    );
+
+    my $modified_method_code = sub {
+        my ( $self, @args ) = @_;
+        my @result;
+        my $result;
+
+        $_->(@_) for @$before;
+
+        my $around_modifier =
+          _compile_around_modifier( \&$super_method, @$around );
+
+        ( defined wantarray )
+          ? (
+            wantarray
+            ? ( @result = $around_modifier->(@_) )
+            : ( $result = $around_modifier->(@_) )
+          )
+          : ( $around_modifier->(@_) );
+
+        $_->(@_) for @$after;
+
+        return unless defined wantarray;
+        return wantarray ? @result : $result;
+    };
+
+    # now bind the new method to the appropriate symbol
+    _bind_coderef_to_symbol( $modified_method_code, $full_method );
+}
+
+# the private method for declaring inheritance, we can here overide the
+# caller class with a random one, useful for our internal cooking, see import().
+sub _extends_class($;$) {
+    my ( $mothers, $class ) = @_;
+    $class = getscope() unless defined $class;
+
+    # then we inherit from all the mothers given, if they are valid
+    foreach my $mother (@$mothers) {
+        confess "Class '$mother' is unknown, cannot extends"
+          unless Coat::Meta->exists($mother);
+        Coat::Meta->extends( $class, $mother );
+    }
+
+    # Add all the mothers to our ancestors.
+    # The extends mechanism overwrite the @ISA list.
+    { no strict 'refs'; @{"${class}::ISA"} = @$mothers; }
+}
+
+sub _bind_coderef_to_symbol($$) {
     my ( $coderef, $symbol ) = @_;
     {
         no strict 'refs';
@@ -276,40 +228,7 @@ sub __bind_coderef_to_symbol($$) {
     }
 }
 
-# check the attributes integrity
-sub __value_is_valid($$) {
-    my ( $value, $type ) = @_;
-    return 1 if $type eq 'Scalar';
 
-    my $lexical_rules = {
-        Int     => '^\d+$',
-        String  => '\w*',
-        Boolean => '^[01]$',
-    };
-
-    if ( defined $lexical_rules->{$type} ) {
-        my $pattern = $lexical_rules->{$type};
-        return $value =~ /$pattern/;
-    }
-
-    # refs
-    elsif ( $type eq 'ArrayRef' ) {
-        return ref($value) eq 'ARRAY';
-    }
-
-    elsif ( $type eq 'HashRef' ) {
-        return ref($value) eq 'HASH';
-    }
-
-    elsif ( $type eq 'CodeRef' ) {
-        return ref($value) eq 'CODE';
-    }
-
-    # take the type as a classname
-    else {
-        return ref($value) eq $type;
-    }
-}
 
 1;
 __END__
@@ -318,7 +237,7 @@ __END__
 
 =head1 NAME
 
-Coat -- A light and self-dependant meta-class for Perl5
+Coat -- A light and self-dependent meta-class for Perl5
 
 =head1 DESCRIPTION
 
@@ -355,8 +274,8 @@ Here is a basic example with a class "Point":
                # inherits from Coat::Object, the mother-class.
 
     # describe attributes...
-    has 'x' => (type => 'Int', default => 0);
-    has 'y' => (type => 'Int', default => 0);
+    has 'x' => (isa => 'Int', default => 0);
+    has 'y' => (isa => 'Int', default => 0);
 
     # and your done
     1;
@@ -369,7 +288,7 @@ Here is a basic example with a class "Point":
 Note that there's no need to import the "strict" and "warnings" modules, it's
 already exported by Coat when you use it.
 
-=head1 STATIC METHODS
+=head1 METHODS
 
 Coat provides you with static methods you use to define your class.
 They're respectingly dedicated to set inheritance, declare attributes
@@ -389,7 +308,7 @@ this documentation:
   use Coat;
   extends 'Point';
 
-  has 'z' => (type => 'Int', default => 0):
+  has 'z' => (isa => 'Int', default => 0):
 
   my $point3d = new Point3D x => 1, y => 3, z => 1;
   $point3d->x;    # will return: 1
@@ -403,21 +322,27 @@ The static method B<has> allows you to define attributes for your class.
 You can handle each attribute options with the %options hashtable. The
 following options are supported:
 
-=head3 type
+=head3 isa
 
-More to come later here...
+When declaring an attribute, it is possible to restrict allowed values 
+to those that validate a type.
+
+See Coat::Types for supported types.
 
 =head3 default
 
 The attribute's default value (the attribute will have this
 value at instanciation time if none given).
 
-=head1 METHOD MODIFIERS (HOOKS)
+Be aware that like with Moose, only plain scalars and code references are allowed 
+when declaring a default value (wrap other references in subs).
+
+=head2 METHOD MODIFIERS (HOOKS)
 
 Like C<Moose>, Coat lets you define hooks. There are three kind of hooks :
 before, after and around.
 
-=head2 before
+=head3 before
 
 When writing a "before" hook you can catch the call to an inherited method,
 and execute some code before the inherited method is called.
@@ -439,7 +364,7 @@ Example:
   };
 
 
-=head2 after
+=head3 after
 
 When writing an "after" hook you can catch the call to an inherited method and 
 execute some code after the original method is executed. You receive in your
@@ -463,7 +388,7 @@ Example:
     $flag = 1;
   };
 
-=head2 around
+=head3 around
 
 When writing an "around" hook you can catch the call to an inherited method and 
 actually redefine it on-the-fly.
